/*
 * Copyright (c) 2014, Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * o Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * o Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * o Neither the name of Freescale Semiconductor, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "bootloader_common.h"
#include "context.h"
#include "property.h"
#include "fsl_device_registers.h"
#include "qspi.h"
#include "fsl_assert.h"

////////////////////////////////////////////////////////////////////////////////
// Declarations
////////////////////////////////////////////////////////////////////////////////
enum
{
    kDefaultClock               = DEFAULT_SYSTEM_CLOCK,
    kFLL_HighFreq               = 96000000ul,
    kHIRC                       = 48000000u,
    kMinCoreClockWithUsbSupport = 20000000u,
    kMaxCoreClock               = 120000000u,
    kMaxBusClock                = 60000000u,
    kMaxFlexBusClock            = 60000000u,
    kMaxFlashClock              = 25000000u,

    kDivider_Min                = 1,
    kDivider_Max                = 16,
};


////////////////////////////////////////////////////////////////////////////////
// Code
////////////////////////////////////////////////////////////////////////////////

// See bootloader_common for documentation on this function.
void configure_clocks()
{
#if defined (BL_TARGET_ROM)

    // 1. Read clock flags and divider from bootloader config in property store
    bootloader_configuration_data_t * config = &g_bootloaderContext.propertyInterface->store->configurationData;
    uint8_t options = config->clockFlags;

    // Check if qspi module needs high FLL frequency
    uint32_t qspiNeedHighFreq;
    quadspi_get_property(kQspiFlashProperty_HighFreq, &qspiNeedHighFreq);

    // Check if the USB HID peripheral is enabled.
    bool isUsbEnabled = config->enabledPeripherals & kPeripheralType_USB_HID;

    // 2. Update SystemCoreClock (DEFAULT_SYSTEM_CLOCK is default value of MCGOUTCLK when OUTDIV1 is equal to 0)
    // In this case, rom should consider the situation that FOPT[LPBOOT] is eqaul to 0
    uint32_t coreClockDivider = ((SIM->CLKDIV1 & SIM_CLKDIV1_OUTDIV1_MASK) >> SIM_CLKDIV1_OUTDIV1_SHIFT) + 1;
    uint32_t mcgOutClock = DEFAULT_SYSTEM_CLOCK;
    SystemCoreClock = mcgOutClock / coreClockDivider;

    // 3. If any of following conditions are met, do nothing (use reset clock config)
    //      (1) NOT High Speed is NOT enabled
    //      (2) usb peripheral is NOT enabled
    //      (3) qspi doesn't need to run at high frequency
    if ((options & kClockFlag_HighSpeed) && !isUsbEnabled && !qspiNeedHighFreq)
    {
        // High speed flag is set (meaning disabled), so just use default clocks.
        return;
    }

    // 4. Update FLL to a high frequency if needed.
    if (qspiNeedHighFreq)
    {
        /* Update clock divider, make sure core clock, bus clock, flexbus clock and flash closk is less than maximum support frequency */
        SIM_CLKDIV1 = SIM_CLKDIV1_OUTDIV1(0x00) |
                      SIM_CLKDIV1_OUTDIV2(0x01) |
                      SIM_CLKDIV1_OUTDIV3(0x03) |
                      SIM_CLKDIV1_OUTDIV4(0x03); /* Update system prescalers */

        /* MCG_C2: RANGE = 2 */
        MCG_C2 |= (uint8_t)MCG_C2_RANGE(0x02);

        /* MCG_C4: DMX32=1,DRST_DRS=3 */
        MCG_C4 |= (uint8_t)((MCG_C4_DMX32_MASK | MCG_C4_DRST_DRS(0x03)));

        while((MCG_S & MCG_S_IREFST_MASK) == 0x00U) { /* Check that the source of the FLL reference clock is the internal reference clock. */
        }
        while((MCG_S & 0x0CU) != 0x00U) {  /* Wait until output of the FLL is selected */
        }

        mcgOutClock = kFLL_HighFreq;
    }

    // 5. Set OUTDIV1 based on divider in config. OUTDIV4 starts out at /1.
    // The divider values are masked by the maximum bits per divider.
    uint32_t div1 = ((~config->clockDivider) & (BM_SIM_CLKDIV1_OUTDIV1 >> BP_SIM_CLKDIV1_OUTDIV1)) + 1;

    // 6. Keep core clock up kMinCoreClockWithUsbSupport if usb is supported.
    uint32_t freq = mcgOutClock;
    // If USB is enabled, the CPU clock must not be allowed to go below 20 MHz
    if (isUsbEnabled)
    {
        while((freq / div1) < kMinCoreClockWithUsbSupport)
        {
            --div1;
        }
    }

    // 7. Keep core clock below kMaxCoreClock
    while ((freq / div1) > kMaxCoreClock)
    {
        ++div1;
    }
    assert((div1 >= kDivider_Min) && (div1 <= kDivider_Max));

    uint32_t div2 = div1;
    uint32_t div3 = div1;
    uint32_t div4 = div1;

    // 8. Keep bus freq below max.
    //
    // The bus clock is divided by OUTDIV2:
    //      MCGOUTCLK -> OUTDIV2 -> bus_clk
    freq = mcgOutClock;
    while ((freq / div2) > kMaxBusClock)
    {
        // Increase bus clock divider.
        ++div2;
    }
    assert((div2 >= kDivider_Min) && (div2 <= kDivider_Max));

    // 9. Keep flex bus freq below max.
    //
    // The flex bus clock is divided by OUTDIV3:
    //     MCGOUTCLK -> OUTDIV3 -> flexbus_clk
    freq = mcgOutClock;
    while ((freq / div3) > kMaxFlexBusClock)
    {
        // Increase flex bus clock divider.
        ++div3;
    }
    assert((div3 >= kDivider_Min) && (div3 <= kDivider_Max));

    // 10. Keep flash freq below max.
    //
    // The flash clock is diveded by OUTDIV4:
    //      MCGOUTCLK -> OUTDIV4 ->flash_clk
    freq = mcgOutClock;
    while ((freq / div4) > kMaxFlashClock)
    {
        // Increase bus/flash clock divider.
        ++div4;
    }

    assert((div4 >= kDivider_Min) && (div4 <= kDivider_Max));

    if((div1 == 1) && ((div2 > 8) || (div3 > 8) || (div4 > 8)))
    {
        return;
    }

    // 11. Now set the dividers
    SIM_CLKDIV1 = SIM_CLKDIV1_OUTDIV1(div1 - 1) |
                  SIM_CLKDIV1_OUTDIV2(div2 - 1) |
                  SIM_CLKDIV1_OUTDIV3(div3 - 1) |
                  SIM_CLKDIV1_OUTDIV4(div4 - 1); /* Update system prescalers */

    // 12. Update SystemCoreClock global.
    SystemCoreClock =  mcgOutClock / div1;

#endif // BL_TARGET_ROM
}

////////////////////////////////////////////////////////////////////////////////
// EOF
////////////////////////////////////////////////////////////////////////////////

