/*
 * Copyright (c) 2013-2014, Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * o Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * o Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * o Neither the name of Freescale Semiconductor, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <string.h>
#include "fsl_platform_common.h" // needed for NULL
#include "fsl_device_registers.h"
#include "qspi.h"
#include "memory.h"
#include "bootloader_common.h"
#include "property.h"
#include "context.h"

////////////////////////////////////////////////////////////////////////////////
// Definitions
////////////////////////////////////////////////////////////////////////////////
qspi_config_t  qspiflash_config_param;



////////////////////////////////////////////////////////////////////////////////
// Code
////////////////////////////////////////////////////////////////////////////////
void rom_memcpy_opt(uint32_t *dst, const uint32_t *src, uint32_t length_in_bytes)
{
    __aeabi_memcpy4((void *)dst, (const void *)src, (size_t) length_in_bytes);
    return;
}


uint32_t rom_memcpy(uint32_t *evt_dst, const uint32_t *evt_src, uint32_t length_in_bytes)
{
    uint16_t * src_16 = NULL , * dst_16 = NULL;
    uint8_t * src_8 = NULL , * dst_8 = NULL;
    uint32_t is_src_align = 0, is_dst_align = 0;

    is_src_align = (uint32_t)evt_src % 4 ;
    is_dst_align = (uint32_t)evt_dst % 4 ;

    if(is_dst_align == 0 && is_src_align == 0)
    {
        /* src and destination are word aligned hence go to word copy routine*/
        /* do opt mem cpy */
        rom_memcpy_opt(evt_dst, evt_src, length_in_bytes);
        return TRUE;
    }
    if(is_dst_align == 2 && is_src_align == 2)
    {
        /* src and destination are not word aligned hence align them to word*/
        /* do half-word copy once */
        dst_16 = (uint16_t*)evt_dst;
        src_16 = (uint16_t *)evt_src;
        *dst_16++ = *src_16++;
        rom_memcpy_opt((uint32_t *)dst_16, (uint32_t *)src_16, length_in_bytes -2);
        return TRUE;
    }
    if(is_dst_align == 3 && is_src_align == 3)
    {
        /* src and destination are not word aligned hence align them to word*/
        /* do byte copy once */
        dst_8 = (uint8_t *)evt_dst;
        src_8 = (uint8_t *)evt_src;
        *dst_8++ = *src_8++;
        rom_memcpy_opt((uint32_t *)dst_8, (uint32_t *)src_8, length_in_bytes - 1);
        return TRUE;
    }
    if(is_dst_align == 1 && is_src_align == 1)
    {
        /* src and destination are not word aligned hence align them to word*/
        /* do byte copy thrice */
        dst_8 = (uint8_t *)evt_dst;
        src_8 = (uint8_t *)evt_src;
        *dst_8++ = *src_8++;
        *dst_8++ = *src_8++;
        *dst_8++ = *src_8++;
        rom_memcpy_opt((uint32_t *)dst_8, (uint32_t *)src_8, length_in_bytes - 3);
        return TRUE;
    }

    /* do byte copy. since source and destination pointers are not evenly aligned */

    __aeabi_memcpy((void*)evt_dst, (const void*)evt_src, (size_t)length_in_bytes);

    return TRUE;
}

status_t quadspi_get_property(qspiflash_property_t whichProperty, uint32_t * value)
{
    if ((value == NULL))
    {
        return kStatus_InvalidArgument;
    }

    switch (whichProperty)
    {
        case kQspiFlashProperty_PageSize:
           *value = qspiflash_config_param.pagesize;
           break;

        case kQspiFlashProperty_TotalFlashSize:
           *value = qspiflash_config_param.sflash_A1_size+qspiflash_config_param.sflash_A2_size+qspiflash_config_param.sflash_B1_size+qspiflash_config_param.sflash_B2_size ;
           break;

        case kQspiFlashProperty_SectorSize:
           *value = qspiflash_config_param.sectorsize;
           break;

        case kQspiFlashProperty_HighFreq:
           *value = (qspiflash_config_param.sclk_freq == 0x10);
           break;

        default:  // catch inputs that are not recognized
            return kStatus_InvalidArgument;
    }

    return kStatus_Success;
}



void quadspi_wait_while_flash_busy()
{
    uint32_t status_value = 0x1;
    while ((status_value & 0x1)==0x1)

    {
        HW_QUADSPI_IPCR_WR(QUADSPI0_BASE,3 << 24);
        while(HW_QUADSPI_SR_RD(QUADSPI0_BASE) &BM_QUADSPI_SR_BUSY);
        status_value = ARDB;
        HW_QUADSPI_FR_WR(QUADSPI0_BASE, 0x10000); /* read complete */

    }
}

status_t quadspi_mem_read(uint32_t address, uint32_t length, uint8_t * buffer)
{
    memcpy((void *)buffer, (void *)address, length);
    return kStatus_Success;
}

status_t quadspi_program(uint32_t address, uint32_t size, const uint8_t * base )
{
  uint32_t *start_address = (unsigned int *)address;
  uint32_t *end_address = (unsigned int *)(address + size);
  uint32_t *page_address = start_address;
  unsigned int *flash_address = (unsigned int *)base;
  int page_size = qspiflash_config_param.pagesize;
  uint32_t  *current_address = start_address;
  unsigned int data_value;
  int i;

  // 1024 offset for spansion

  HW_QUADSPI_SFAR_WR(QUADSPI0_BASE, (unsigned int)flash_address);
  HW_QUADSPI_MCR_SET(QUADSPI0_BASE, BM_QUADSPI_MCR_CLR_TXF);


  page_address = start_address + (page_size >> 2);

  while (page_address <= end_address)
  {

    // write enable
    HW_QUADSPI_IPCR_SET(QUADSPI0_BASE,1 << 24);
    while(HW_QUADSPI_SR_RD(QUADSPI0_BASE) &BM_QUADSPI_SR_BUSY );
    HW_QUADSPI_MCR_SET(QUADSPI0_BASE, BM_QUADSPI_MCR_CLR_TXF);

    while (current_address < page_address)
    {
      // fill tx fifo (64 bytes)
      for (i = 0; i < 16; i++)
      {
        data_value=*(current_address++);
        HW_QUADSPI_TBDR_WR(QUADSPI0_BASE, byte_swap32(data_value));
      }

      HW_QUADSPI_IPCR_WR(QUADSPI0_BASE,(4 << 24) | page_size); //page program 256bytes
      for (i = 0; i < 48; i++)
      {
        while(HW_QUADSPI_SR_RD(QUADSPI0_BASE) &BM_QUADSPI_SR_TXFULL);
        data_value=*(current_address++);
        HW_QUADSPI_TBDR_WR(QUADSPI0_BASE,byte_swap32(data_value));
      }
      while(HW_QUADSPI_SR_RD(QUADSPI0_BASE) & BM_QUADSPI_SR_BUSY);
      quadspi_wait_while_flash_busy();


    }
    page_address += (page_size >> 2);
    flash_address += (page_size >> 2);
    HW_QUADSPI_SFAR_WR(QUADSPI0_BASE,(unsigned int)flash_address);
  }

  return kStatus_Success;
}



void quadspi_erase_all()
{
    HW_QUADSPI_SFAR_WR(QUADSPI0_BASE , FLASH_BASE_ADDR);

    //write enable
    HW_QUADSPI_IPCR_WR(QUADSPI0_BASE,1 << 24);
    while(HW_QUADSPI_SR_RD(QUADSPI0_BASE) & BM_QUADSPI_SR_BUSY);

    //send erase command
    HW_QUADSPI_IPCR_WR(QUADSPI0_BASE,2 << 24);
    while(HW_QUADSPI_SR_RD(QUADSPI0_BASE) &  BM_QUADSPI_SR_BUSY);

    quadspi_wait_while_flash_busy();

}


void quadspi_erase_sector(uint32_t address)
{

    HW_QUADSPI_SFAR_WR(QUADSPI0_BASE , address);

    //write enable
    HW_QUADSPI_IPCR_WR(QUADSPI0_BASE,1 << 24);
    while(HW_QUADSPI_SR_RD(QUADSPI0_BASE) & BM_QUADSPI_SR_BUSY);

    //send erase sector command, seq 7
    HW_QUADSPI_IPCR_WR(QUADSPI0_BASE,7 << 24);
    while(HW_QUADSPI_SR_RD(QUADSPI0_BASE) &  BM_QUADSPI_SR_BUSY);
    
    quadspi_wait_while_flash_busy();
    

}

void config_ahb_buffer()
{
    /* Configure BUFFER 3 for All Master Access */
    HW_QUADSPI_BUF3CR_SET(QUADSPI0_BASE, BM_QUADSPI_BUF3CR_ALLMST);

    /* If CPU masterID is zero then qspi data is fetched from AHB buffer 0.
    Where as we expect the data to be on BUF 3. Hence we program the Master ID
    for Buffer control register 0,1,2 with a dummy masterIDs
    */
    HW_QUADSPI_BUF0CR_SET(QUADSPI0_BASE, AHB_BFCR_DUMMY_MASTER_ID);
    HW_QUADSPI_BUF1CR_SET(QUADSPI0_BASE, AHB_BFCR_DUMMY_MASTER_ID);
    HW_QUADSPI_BUF2CR_SET(QUADSPI0_BASE, AHB_BFCR_DUMMY_MASTER_ID);
    /*Program the Sequence ID for read operation. */
    /* Get size verified by the designer that A size of 1 written in BF0CR_ADATSZ is how many bytes*/
    /* Set Prefetch Size to 0xFF -> 255*8 bytes to be prefetched */
    HW_QUADSPI_BUF3CR_CLR(QUADSPI0_BASE, BM_QUADSPI_BUF3CR_ADATSZ);
    HW_QUADSPI_BUF3CR_SET(QUADSPI0_BASE, 32<<BP_QUADSPI_BUF3CR_ADATSZ );

    return;

}


static void quadspi_clock_enable()
{
//    HW_SIM_SOPT3_SET(SIM_BASE, SIM_SOPT3_QSPIEN_MASK);
    HW_SIM_SCGC2_SET(SIM_BASE, SIM_SCGC2_QSPI_MASK);
}


static void quadspi_serial_clock_disable()
{

   HW_QUADSPI_MCR_SET(QUADSPI0_BASE,BM_QUADSPI_MCR_MDIS);//disable qspi module
}

static void quadspi_serial_clock_configure()
{
    HW_QUADSPI_SOCCR_CLR(QUADSPI0_BASE, 0x7);
    HW_QUADSPI_SOCCR_SET(QUADSPI0_BASE, 0x4);  //choose IRC48M as the clock source for qspi

}

status_t get_quadspi_config_data()
{

    //Set GPIO for basic Read
    quadspi_iomux_configuration(kQspiType_dual, kQspiMode_non_parallel);
    // Do minimum configuration to run basic read command 0x03
    /* Enable QSPI Module Clock */
    quadspi_clock_enable();
    //TBD:Set clock for Basic Read 1. Enable QuadSPI clock. 2. Enable SCLK
    quadspi_serial_clock_configure();

    /* initialization for AHB Buffer Access */
    config_ahb_buffer();

    /*Enable QuadSPI module */
    HW_QUADSPI_MCR_CLR(QUADSPI0_BASE, BM_QUADSPI_MCR_MDIS);

    property_store_t * propertyStore = g_bootloaderContext.propertyInterface->store;
    
    if (propertyStore->configurationData.qspiConfigPointer == 0xffffffff)
        return kStatus_Fail;

    (void) memcpy((uint32_t*)&qspiflash_config_param, (void *)propertyStore->configurationData.qspiConfigPointer,
        sizeof(qspiflash_config_param));

    if((qspiflash_config_param.sflash_A1_size ==0)&&(qspiflash_config_param.sflash_A2_size ==0)&&(qspiflash_config_param.sflash_B1_size ==0)&&(qspiflash_config_param.sflash_B2_size ==0))
        return kStatus_Fail;
    else
        return kStatus_Success;


}

void configure_quad_spi_controller()
{


    uint32_t smpr;

    /* Reset TX/RX FIFO/BUFFER */
    HW_QUADSPI_MCR_SET(QUADSPI0_BASE, BM_QUADSPI_MCR_CLR_TXF);
    HW_QUADSPI_MCR_SET(QUADSPI0_BASE, BM_QUADSPI_MCR_CLR_RXF);

    /* Enable DDR mode*/
    if (qspiflash_config_param.ddr_mode_enable)
    {
        HW_QUADSPI_MCR_SET(QUADSPI0_BASE, BM_QUADSPI_MCR_DDR_EN);
    }

    if (qspiflash_config_param.dqs_enable)
    {
        HW_QUADSPI_SOCCR_SET(QUADSPI0_BASE, 0x100);
        /* Set DQS LoopBack to enable dummy pad if enabled from Configuration */
        if(qspiflash_config_param.dqs_loopback)
        {
        HW_QUADSPI_SOCCR_SET(QUADSPI0_BASE, 0x200);
        }
        else
        {
        HW_QUADSPI_SOCCR_CLR(QUADSPI0_BASE, 0x200);
        }
    }

    /* Program Sampling Parameters */
    smpr = HW_QUADSPI_SMPR_RD(QUADSPI0_BASE);
    smpr &= ~(BM_QUADSPI_SMPR_DDRSMP|BM_QUADSPI_SMPR_FSDLY | BM_QUADSPI_SMPR_FSPHS );
    smpr |= (((qspiflash_config_param.fsdly &BM_QUADSPI_SMPR_FSDLY) << BP_QUADSPI_SMPR_FSDLY) |
    ((qspiflash_config_param.fsphs &  BM_QUADSPI_SMPR_FSPHS) <<  BP_QUADSPI_SMPR_FSPHS) |
    ((qspiflash_config_param.ddrsmp & BM_QUADSPI_SMPR_DDRSMP) << BP_QUADSPI_SMPR_DDRSMP));
    HW_QUADSPI_SMPR_WR(QUADSPI0_BASE, smpr);

    /* Enable parralel flash access */
    if (qspiflash_config_param.parallel_mode_enable)
    {
        HW_QUADSPI_BFGENCR_SET(QUADSPI0_BASE, BM_QUADSPI_BFGENCR_PAR_EN);
    }

    /* Setup SFLASH hold and setup time */
    HW_QUADSPI_FLSHCR_SET(QUADSPI0_BASE, (qspiflash_config_param.cs_hold_time) << BP_QUADSPI_FLSHCR_TCSH);
    HW_QUADSPI_FLSHCR_SET(QUADSPI0_BASE, (qspiflash_config_param.cs_setup_time) << BP_QUADSPI_FLSHCR_TCSS);

    HW_QUADSPI_SFA1AD_WR(QUADSPI0_BASE, ((uint32_t)quadspi0_base_address + (uint32_t)qspiflash_config_param.sflash_A1_size));
    HW_QUADSPI_SFA2AD_WR(QUADSPI0_BASE,  (HW_QUADSPI_SFA1AD_RD(QUADSPI0_BASE) + (uint32_t)qspiflash_config_param.sflash_A2_size));
    HW_QUADSPI_SFB1AD_WR(QUADSPI0_BASE,  (HW_QUADSPI_SFA2AD_RD(QUADSPI0_BASE) + (uint32_t)qspiflash_config_param.sflash_B1_size));
    HW_QUADSPI_SFB2AD_WR(QUADSPI0_BASE,  (HW_QUADSPI_SFB1AD_RD(QUADSPI0_BASE) + (uint32_t)qspiflash_config_param.sflash_B2_size));


    /* Configure and Enable SCLK*/
    quadspi_serial_clock_configure();

    /*Enable QuadSPI module */
    HW_QUADSPI_MCR_CLR(QUADSPI0_BASE, BM_QUADSPI_MCR_MDIS);
    return;

}

void configure_lut()
{
    uint32_t temp;
    uint32_t cnt;
    uint32_t look_up_table_seq = QSPI_READ_SEQ_ID;

    for (cnt = 0;cnt < COMAND_SEQ_LEN; cnt++)
    {
      temp = (qspiflash_config_param.command_seq[cnt] | ( (uint32_t)qspiflash_config_param.command_seq[cnt + 1] << 16));
      cnt++;

      HW_QUADSPI_LUTn_WR(QUADSPI0_BASE,look_up_table_seq++, temp);
    }

    return;
}


static void configure_quadspi()
{
    /* Disable QSPI Module */

    HW_QUADSPI_MCR_SET(QUADSPI0_BASE,BM_QUADSPI_MCR_MDIS);


    /* Disable QSPI Serial Clock */
    quadspi_serial_clock_disable();

    /* QUADSPI IOMUX config */
    quadspi_iomux_configuration((qspiflash_type)qspiflash_config_param.sflash_type,(qspiflash_is_parallel)qspiflash_config_param.parallel_mode_enable);

    /*Program the look up table with flash sequence command*/
    configure_lut();

    /*Configure QuadSPI controller*/
    configure_quad_spi_controller();

    return;

}

status_t quadspi_init(void)
{
    uint32_t status;
    status = get_quadspi_config_data();
    if (status == kStatus_Fail) return kStatus_Fail;
    configure_quadspi();
    return kStatus_Success;
}

