/*
 * Copyright (c) 2014, Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * o Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * o Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * o Neither the name of Freescale Semiconductor, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include "fsl_platform_common.h" // needed for NULL
#include "fsl_device_registers.h"
#include "qspi.h"
#include "memory.h"
#include "bootloader_common.h"
#include "property.h"
#include "context.h"
#include "fsl_assert.h"

////////////////////////////////////////////////////////////////////////////////
// Definitions
////////////////////////////////////////////////////////////////////////////////
enum
{
    //! @brief Page size of simulated qspi flash
    kQspiFlash_PageSize             = 256u,
    //! @brief Sector size of simulated qspi flash
    kQspiFlash_SectorSize           = 1024u,
    //! @brief Total size of simulated qspi flash
    kQspiFlash_TotalFlashSize       = 65536u,

    //! @brief Actual start address of simulated qspi flash
    kQspiFlash_ActualStart          = 0x20010000u,
    //! @brief Base address of real qspi flash
    kQspiFlash_BaseAddress          = 0x68000000u,
};


////////////////////////////////////////////////////////////////////////////////
// Declarations
////////////////////////////////////////////////////////////////////////////////
//! @brief Return actual address mapped with given address
uint32_t qspi_sim_get_actual_address(uint32_t address);


////////////////////////////////////////////////////////////////////////////////
// Code
////////////////////////////////////////////////////////////////////////////////

// See qspi.h for documenation on this function
status_t quadspi_get_property(qspiflash_property_t whichProperty, uint32_t * value)
{
    if ((value == NULL))
    {
        return kStatus_InvalidArgument;
    }

    switch (whichProperty)
    {
    case kQspiFlashProperty_PageSize:
        *value = kQspiFlash_PageSize;
        break;

    case kQspiFlashProperty_TotalFlashSize:
        *value = kQspiFlash_TotalFlashSize;
        break;

    case kQspiFlashProperty_SectorSize:
        *value = kQspiFlash_SectorSize;
        break;

    case kQspiFlashProperty_HighFreq:
        *value = false;
        break;

    default:  // catch inputs that are not recognized
        return kStatus_InvalidArgument;
    }

    return kStatus_Success;
}

// See qspi.h for documenation on this function
status_t quadspi_program(uint32_t address, uint32_t size, const uint8_t * base )
{
    uint32_t actualAddress = qspi_sim_get_actual_address(address);

    // Check if the page to be programmed is empty.
    uint32_t *pDstLongWord = (uint32_t *)actualAddress;
    for(uint32_t i=0; i< kQspiFlash_PageSize; i+= 4)
    {
        if( *pDstLongWord != 0xFFFFFFFFUL)
        {
            return kStatus_QspiFlashCommandFailure;
        }
        ++pDstLongWord;
    }

    memcpy((void*)actualAddress, base, size);

    return kStatus_Success;
}

// See qspi.h for documenation on this function
status_t quadspi_init(void)
{
    return kStatus_Success;
}

// See qspi.h for documenation on this function
void quadspi_erase_all()
{
    memset((void*)kQspiFlash_ActualStart, 0xFF, kQspiFlash_TotalFlashSize);
}


// See qspi.h for documenation on this function
void quadspi_erase_sector(uint32_t address)
{
    uint32_t actualAddress = qspi_sim_get_actual_address(address);

    memset((void*)actualAddress, 0xFF, kQspiFlash_SectorSize);
}

uint32_t qspi_sim_get_actual_address(uint32_t address)
{
    assert ((address >= kQspiFlash_BaseAddress) && (address < kQspiFlash_BaseAddress + kQspiFlash_TotalFlashSize));

    uint32_t offset =  address - kQspiFlash_BaseAddress;

    uint32_t actualAddress = kQspiFlash_ActualStart + offset;

    return actualAddress;
}

